# Brute force coherence (Gabriele Vajente, 2016-10-18)
# LIGO-centric functions

import os
import subprocess
import numpy

try:
    from urllib.parse import urlparse
except ImportError:  # python < 3
    from urlparse import urlparse

import gwdatafind

try:
    import framel
    use_framel = True
except:
    from gwpy.timeseries import TimeSeries
    use_framel = False

# global variables for LIGO data I/O
files = []

# wrapper around the LIGO function to find where data is, returns a list of files
def find_data_path(observatory, gpsb, gpse):
    obs = observatory[0]
    urls = gwdatafind.find_urls(
        obs,
        "{}1_R".format(obs),
        int(gpsb),
        int(gpse),
        urltype='file',
    )
    return [str(urlparse(x).path) for x in urls]

# get the list of channels
def get_channel_list(opt, gpsb):
    
    files0 = find_data_path(opt.ifo, gpsb, gpsb+1)
    if not files0:
        ermsg = 'No frame files found of type {:s}1_R for ' \
                'time {:d}'.format(opt.ifo[0], int(gpsb))
        raise AttributeError(ermsg)
    os.system('/usr/bin/FrChannels ' + files0[0] + ' > bruco.channels')
    f = open('bruco.channels')
    lines = f.readlines()
    channels = []
    sample_rate = []
    for l in lines:
        ll = l.split()
        if ll[0][1] != '0':
            # remove all L0/H0 channels
            channels.append(ll[0])
            sample_rate.append(int(ll[1]))
    channels = numpy.array(channels)
    sample_rate = numpy.array(sample_rate)
    # remove temporary channel list
    os.system('rm bruco.channels')
    
    return channels, sample_rate
    
if use_framel:
    def getRawData(channel, gps, dt):
        """Read data from RAW file (using pylal):
        ch  = channel name
        gps = gps time
        dt  = duration
        """
    
        global files
    
        # get the list of frame files
        obs = channel[0:2]
        if len(files) == 0:
            files = find_data_path(obs, gps, gps+dt)
        # read data from all files
        data = numpy.array([]) 
        for f in files:
            # get the frame file start GPS and duration from the name 
            gps0 = int(f.split('-')[-2])
            gps1 = gps0 + int(f.split('-')[-1].split('.')[-2])
            # find the right time segment to load from this file
            gps0 = max(gps0, gps)
            gps1 = min(gps1, gps + dt)
            # read data and append
            x = framel.frgetvect(f, channel, gps0, gps1-gps0)
            data = numpy.concatenate([data, x[0]])
        return data, int(1/x[3][0])
else:
    def getRawData(channel, gps, dt, source=None):
        """Read data from RAW file (using gwpy):
        ch  = channel name
        gps = gps time
        dt  = duration
        """

        global files

        # get the list of frame files
        obs = channel[0:2]
        if source:
            files_ = open(source, 'r').readlines()
            files_ = [f[:-1] for f in files_]
            print('Reading %s from' % channel)
            print(files_)
            data = TimeSeries.read(files_, channel, start=gps, end=gps+dt)
        else:
            if len(files) == 0:
                files = find_data_path(obs, gps, gps+dt)
            data = TimeSeries.read(files, channel, start=gps, end=gps+dt, format='gwf.framecpp', type='adc')

        return data.value, int(data.sample_rate.to('Hz').value)



#!/usr/bin/env python
"""Module providing Multirate signal processing functionality.
Largely based on MATLAB's Multirate signal processing toolbox with consultation 
of Octave m-file source code.
"""

from __future__ import print_function

import sys
import fractions
import numpy
from scipy import signal


def downsample(s, n, phase=0):
    """Decrease sampling rate by integer factor n with included offset phase.
    """
    return s[phase::n]


def upsample(s, n, phase=0):
    """Increase sampling rate by integer factor n  with included offset phase.
    """
    return numpy.roll(numpy.kron(s, numpy.r_[1, numpy.zeros(n-1)]), phase)


def decimate(s, r, n=None, fir=False):
    """Decimation - decrease sampling rate by r. The decimation process filters 
    the input data s with an order n lowpass filter and then resamples the 
    resulting smoothed signal at a lower rate. By default, decimate employs an 
    eighth-order lowpass Chebyshev Type I filter with a cutoff frequency of 
    0.8/r. It filters the input sequence in both the forward and reverse 
    directions to remove all phase distortion, effectively doubling the filter 
    order. If 'fir' is set to True decimate uses an order 30 FIR filter (by 
    default otherwise n), instead of the Chebyshev IIR filter. Here decimate 
    filters the input sequence in only one direction. This technique conserves 
    memory and is useful for working with long sequences.
    """
    if fir:
        if n is None:
            n = 30
        b = signal.firwin(n, 1.0/r)
        a = 1
        f = signal.lfilter(b, a, s)
    else: #iir
        if n is None:
            n = 8
        b, a = signal.cheby1(n, 0.05, 0.8/r)
        f = signal.filtfilt(b, a, s)
    return downsample(f, r)


def interp(s, r, l=4, alpha=0.5):
    """Interpolation - increase sampling rate by integer factor r. Interpolation 
    increases the original sampling rate for a sequence to a higher rate. interp
    performs lowpass interpolation by inserting zeros into the original sequence
    and then applying a special lowpass filter. l specifies the filter length 
    and alpha the cut-off frequency. The length of the FIR lowpass interpolating
    filter is 2*l*r+1. The number of original sample values used for 
    interpolation is 2*l. Ordinarily, l should be less than or equal to 10. The 
    original signal is assumed to be band limited with normalized cutoff 
    frequency 0=alpha=1, where 1 is half the original sampling frequency (the 
    Nyquist frequency). The default value for l is 4 and the default value for 
    alpha is 0.5.
    """
    b = signal.firwin(2*l*r+1, alpha/r);
    a = 1
    return r*signal.lfilter(b, a, upsample(s, r))[r*l+1:-1]


def resample(s, p, q, h=None):
    """Change sampling rate by rational factor. This implementation is based on
    the Octave implementation of the resample function. It designs the 
    anti-aliasing filter using the window approach applying a Kaiser window with
    the beta term calculated as specified by [2].
    
    Ref [1] J. G. Proakis and D. G. Manolakis,
    Digital Signal Processing: Principles, Algorithms, and Applications,
    4th ed., Prentice Hall, 2007. Chap. 6

    Ref [2] A. V. Oppenheim, R. W. Schafer and J. R. Buck, 
    Discrete-time signal processing, Signal processing series,
    Prentice-Hall, 1999
    """
    gcd = fractions.gcd(p,q)
    if gcd>1:
        p=p/gcd
        q=q/gcd
    
    if h is None: #design filter
        #properties of the antialiasing filter
        log10_rejection = -3.0
        stopband_cutoff_f = 1.0/(2.0 * max(p,q))
        roll_off_width = stopband_cutoff_f / 10.0
    
        #determine filter length
        #use empirical formula from [2] Chap 7, Eq. (7.63) p 476
        rejection_db = -20.0*log10_rejection;
        l = numpy.ceil((rejection_db-8.0) / (28.714 * roll_off_width))
  
        #ideal sinc filter
        t = numpy.arange(-l, l + 1)
        ideal_filter=2*p*stopband_cutoff_f*numpy.sinc(2*stopband_cutoff_f*t)  
  
        #determine parameter of Kaiser window
        #use empirical formula from [2] Chap 7, Eq. (7.62) p 474
        beta = signal.kaiser_beta(rejection_db)
          
        #apodize ideal filter response
        h = numpy.kaiser(2*l+1, beta)*ideal_filter

    ls = s.shape[0]
    lh = h.shape[0]

    l = (lh - 1)/2.0
    ly = int(numpy.ceil(ls*p/float(q)))

    #pre and postpad filter response
    nz_pre = int(numpy.floor(q - numpy.mod(l,q)))
    hpad = h[-lh+nz_pre:]

    offset = int(numpy.floor((l+nz_pre)/q))
    nz_post = 0;
    while numpy.ceil(((ls-1)*p + nz_pre + lh + nz_post )/q ) - offset < ly:
        nz_post += 1
    hpad = hpad[:lh + nz_pre + nz_post]

    #filtering
    xfilt = upfirdn(s, hpad, p, q)

    return xfilt[offset-1:offset-1+ly]


def upfirdn(s, h, p, q):
    """Upsample signal s by p, apply FIR filter as specified by h, and 
    downsample by q. Using fftconvolve as opposed to lfilter as it does not seem
    to do a full convolution operation (and its much faster than convolve).
    """
    return downsample(signal.fftconvolve(h, upsample(s, p)), q)

def main():
    """Show simple use cases for functionality provided by this module. Each 
    example below attempts to mimic the examples provided by mathworks MATLAB
    documentation, http://www.mathworks.com/help/toolbox/signal/
    """
    import pylab
    argv = sys.argv
    if len(argv) != 1:
        print('usage: python -m pim.sp.multirate', file=sys.stderr)
        sys.exit(2)

    #Downsample
    x = numpy.arange(1, 11)
    print('Down Sampling %s by 3' % x)
    print( downsample(x, 3))
    print('Down Sampling %s by 3 with phase offset 2' % x)
    print( downsample(x, 3, phase=2))

    #Upsample
    x = numpy.arange(1, 5)
    print('Up Sampling %s by 3' % x)
    print(upsample(x, 3))
    print('Up Sampling %s by 3 with phase offset 2' % x)
    print(upsample(x, 3, 2))

    #Decimate
    t = numpy.arange(0, 1, 0.00025)
    x = numpy.sin(2*numpy.pi*30*t) + numpy.sin(2*numpy.pi*60*t)
    y = decimate(x,4)
    pylab.figure()
    pylab.subplot(2, 1, 1)
    pylab.title('Original Signal')
    pylab.stem(numpy.arange(len(x[0:120])), x[0:120])
    pylab.subplot(2, 1, 2)
    pylab.title('Decimated Signal')
    pylab.stem(numpy.arange(len(y[0:30])), y[0:30])

    #Interp
    t = numpy.arange(0, 1, 0.001)
    x = numpy.sin(2*numpy.pi*30*t) + numpy.sin(2*numpy.pi*60*t)
    y = interp(x,4)
    pylab.figure()
    pylab.subplot(2, 1, 1)
    pylab.title('Original Signal')
    pylab.stem(numpy.arange(len(x[0:30])), x[0:30])
    pylab.subplot(2, 1, 2)
    pylab.title('Interpolated Signal')
    pylab.stem(numpy.arange(len(y[0:120])), y[0:120])

    #upfirdn
    L = 147.0 
    M = 160.0
    N = 24.0*L
    h = signal.firwin(N-1, 1/M, window=('kaiser', 7.8562))
    h = L*h
    Fs = 48000.0
    n = numpy.arange(0, 10239)
    x  = numpy.sin(2*numpy.pi*1000/Fs*n)
    y = upfirdn(x, h, L, M)
    pylab.figure()
    pylab.stem(n[1:49]/Fs, x[1:49])
    pylab.stem(n[1:45]/(Fs*L/M), y[13:57], 'r', markerfmt='ro',)
    pylab.xlabel('Time (sec)')
    pylab.ylabel('Signal value')

    #resample
    fs1 = 10.0
    t1 = numpy.arange(0, 1 + 1.0/fs1, 1.0/fs1)
    x = t1
    y = resample(x, 3, 2)
    t2 = numpy.arange(0,(len(y)))*2.0/(3.0*fs1)
    pylab.figure()
    pylab.plot(t1, x, '*')
    pylab.plot(t2, y, 'o')
    pylab.plot(numpy.arange(-0.5,1.5, 0.01), numpy.arange(-0.5,1.5, 0.01), ':')
    pylab.legend(('original','resampled'))
    pylab.xlabel('Time')
    
    x = numpy.hstack([numpy.arange(1,11), numpy.arange(9,0,-1)])
    y = resample(x,3,2)
    pylab.figure()
    pylab.subplot(2, 1, 1)
    pylab.title('Edge Effects Not Noticeable')
    pylab.plot(numpy.arange(19)+1, x, '*')
    pylab.plot(numpy.arange(29)*2/3.0 + 1, y, 'o')
    pylab.legend(('original', 'resampled'))
    x = numpy.hstack([numpy.arange(10, 0, -1), numpy.arange(2,11)])
    y = resample(x,3,2)
    pylab.subplot(2, 1, 2)
    pylab.plot(numpy.arange(19)+1, x, '*')
    pylab.plot(numpy.arange(29)*2/3.0 + 1, y, 'o')
    pylab.title('Edge Effects Very Noticeable')
    pylab.legend(('original', 'resampled'))

    pylab.show()
    return 0

if __name__ == '__main__':
    sys.exit(main())



# Brute force coherence (Gabriele Vajente, 2017-11-17)
# 40m-centric functions

import os
#from pylal import frutils, Fr
import subprocess
from pylab import *
import numpy
import nds2

# global variables for LIGO data I/O
files = []

# get the list of channels
def get_channel_list(opt, gpsb):
    
    conn = nds2.connection('nds40.ligo.caltech.edu')
    ch = conn.find_channels('C1:*DQ')
    channels = [c.name for c in ch]
    sample_rate = [c.sample_rate for c in ch]
   
    return array(channels), array(sample_rate)
    
def getRawData(channel, gps, dt):
    """Read data from RAW file:
    ch  = channel name
    gps = gps time
    dt  = duration
    """

    conn = nds2.connection('nds40.ligo.caltech.edu')
    buf = conn.fetch(gps, gps+dt, [channel])
    return buf[0].data, int(buf[0].length/dt)




# This code is in the public domain, it comes
# with absolutely no warranty and you can do
# absolutely whatever you want with it.

__date__ = '16 March 2015'
__version__ = '1.10'
__doc__= """
This is markup.py - a Python module that attempts to
make it easier to generate HTML/XML from a Python program
in an intuitive, lightweight, customizable and pythonic way.
It works with both python 2 and 3.

The code is in the public domain.

Version: %s as of %s.

Documentation and further info is at http://markup.sourceforge.net/

Please send bug reports, feature requests, enhancement
ideas or questions to nogradi at gmail dot com.

Installation: drop markup.py somewhere into your Python path.
""" % ( __version__, __date__ )

try:
    basestring
    import string
except:
    # python 3
    basestring = str
    string = str
    long = int

# tags which are reserved python keywords will be referred 
# to by a leading underscore otherwise we end up with a syntax error
import keyword

class element:
    """This class handles the addition of a new element."""

    def __init__( self, tag, case='lower', parent=None ):
        self.parent = parent

        if case == 'upper':
            self.tag = tag.upper( )
        elif case == 'lower':
            self.tag = tag.lower( )
        elif case =='given':
            self.tag = tag
        else:
            self.tag = tag
    
    def __call__( self, *args, **kwargs ):
        if len( args ) > 1:
            raise ArgumentError( self.tag )

        # if class_ was defined in parent it should be added to every element
        if self.parent is not None and self.parent.class_ is not None:
            if 'class_' not in kwargs:
                kwargs['class_'] = self.parent.class_
            
        if self.parent is None and len( args ) == 1:
            x = [ self.render( self.tag, False, myarg, mydict ) for myarg, mydict in _argsdicts( args, kwargs ) ]
            return '\n'.join( x )
        elif self.parent is None and len( args ) == 0:
            x = [ self.render( self.tag, True, myarg, mydict ) for myarg, mydict in _argsdicts( args, kwargs ) ]
            return '\n'.join( x )
            
        if self.tag in self.parent.twotags:
            for myarg, mydict in _argsdicts( args, kwargs ):
                self.render( self.tag, False, myarg, mydict )
        elif self.tag in self.parent.onetags:
            if len( args ) == 0:
                for myarg, mydict in _argsdicts( args, kwargs ):
                    self.render( self.tag, True, myarg, mydict )    # here myarg is always None, because len( args ) = 0
            else:
                raise ClosingError( self.tag )
        elif self.parent.mode == 'strict_html' and self.tag in self.parent.deptags:
            raise DeprecationError( self.tag )
        else:
            raise InvalidElementError( self.tag, self.parent.mode )
    
    def render( self, tag, single, between, kwargs ):
        """Append the actual tags to content."""

        out = "<%s" % tag
        for key, value in list( kwargs.items( ) ):
            if value is not None:               # when value is None that means stuff like <... checked>
                key = key.strip('_')            # strip this so class_ will mean class, etc.
                if key == 'http_equiv':         # special cases, maybe change _ to - overall?
                    key = 'http-equiv'
                elif key == 'accept_charset':
                    key = 'accept-charset'
                out = "%s %s=\"%s\"" % ( out, key, escape( value ) )
            else:
                out = "%s %s" % ( out, key )
        if between is not None:
            out = "%s>%s</%s>" % ( out, between, tag )
        else:
            if single:
                out = "%s />" % out
            else:
                out = "%s>" % out
        if self.parent is not None:
            self.parent.content.append( out )
        else:
            return out
    
    def close( self ):
        """Append a closing tag unless element has only opening tag."""

        if self.tag in self.parent.twotags:
            self.parent.content.append( "</%s>" % self.tag )
        elif self.tag in self.parent.onetags:
            raise ClosingError( self.tag )
        elif self.parent.mode == 'strict_html' and self.tag in self.parent.deptags:
            raise DeprecationError( self.tag )

    def open( self, **kwargs ):
        """Append an opening tag."""

        if self.tag in self.parent.twotags or self.tag in self.parent.onetags:
            self.render( self.tag, False, None, kwargs )
        elif self.mode == 'strict_html' and self.tag in self.parent.deptags:
            raise DeprecationError( self.tag )

class page:
    """This is our main class representing a document. Elements are added
    as attributes of an instance of this class."""

    def __init__( self, mode='strict_html', case='lower', onetags=None, twotags=None, separator='\n', class_=None ):
        """Stuff that effects the whole document.

        mode -- 'strict_html'   for HTML 4.01 (default)
                'html'          alias for 'strict_html'
                'loose_html'    to allow some deprecated elements
                'xml'           to allow arbitrary elements

        case -- 'lower'         element names will be printed in lower case (default)
                'upper'         they will be printed in upper case
                'given'         element names will be printed as they are given

        onetags --              list or tuple of valid elements with opening tags only
        twotags --              list or tuple of valid elements with both opening and closing tags
                                these two keyword arguments may be used to select
                                the set of valid elements in 'xml' mode
                                invalid elements will raise appropriate exceptions
        
        separator --            string to place between added elements, defaults to newline
        
        class_ --               a class that will be added to every element if defined"""
        
        valid_onetags = [ "AREA", "BASE", "BR", "COL", "FRAME", "HR", "IMG", "INPUT", "LINK", "META", "PARAM" ]
        valid_twotags = [ "A", "ABBR", "ACRONYM", "ADDRESS", "B", "BDO", "BIG", "BLOCKQUOTE", "BODY", "BUTTON",
                "CAPTION", "CITE", "CODE", "COLGROUP", "DD", "DEL", "DFN", "DIV", "DL", "DT", "EM", "FIELDSET",
                "FORM", "FRAMESET", "H1", "H2", "H3", "H4", "H5", "H6", "HEAD", "HTML", "I", "IFRAME", "INS",
                "KBD", "LABEL", "LEGEND", "LI", "MAP", "NOFRAMES", "NOSCRIPT", "OBJECT", "OL", "OPTGROUP",
                "OPTION", "P", "PRE", "Q", "SAMP", "SCRIPT", "SELECT", "SMALL", "SPAN", "STRONG", "STYLE",
                "SUB", "SUP", "TABLE", "TBODY", "TD", "TEXTAREA", "TFOOT", "TH", "THEAD", "TITLE", "TR",
                "TT", "UL", "VAR" ]
        deprecated_onetags = [ "BASEFONT", "ISINDEX" ]
        deprecated_twotags = [ "APPLET", "CENTER", "DIR", "FONT", "MENU", "S", "STRIKE", "U" ]

        self.header = [ ]
        self.content = [ ]
        self.footer = [ ]
        self.case = case
        self.separator = separator

        # init( ) sets it to True so we know that </body></html> has to be printed at the end
        self._full = False
        self.class_= class_

        if mode == 'strict_html' or mode == 'html':
            self.onetags = valid_onetags
            self.onetags += list( map( string.lower, self.onetags ) )
            self.twotags = valid_twotags
            self.twotags += list( map( string.lower, self.twotags ) )
            self.deptags = deprecated_onetags + deprecated_twotags
            self.deptags += list( map( string.lower, self.deptags ) )
            self.mode = 'strict_html'
        elif mode == 'loose_html':
            self.onetags = valid_onetags + deprecated_onetags 
            self.onetags += list( map( string.lower, self.onetags ) )
            self.twotags = valid_twotags + deprecated_twotags
            self.twotags += list( map( string.lower, self.twotags ) )
            self.mode = mode
        elif mode == 'xml':
            if onetags and twotags:
                self.onetags = onetags
                self.twotags = twotags
            elif ( onetags and not twotags ) or ( twotags and not onetags ):
                raise CustomizationError( )
            else:
                self.onetags = russell( )
                self.twotags = russell( )
            self.mode = mode
        else:
            raise ModeError( mode )

    def __getattr__( self, attr ):

        # tags should start with double underscore
        if attr.startswith("__") and attr.endswith("__"):
            raise AttributeError( attr )
        # tag with single underscore should be a reserved keyword
        if attr.startswith( '_' ):
            attr = attr.lstrip( '_' ) 
            if attr not in keyword.kwlist:
                raise AttributeError( attr )

        return element( attr, case=self.case, parent=self )

    def __str__( self ):
        
        if self._full and ( self.mode == 'strict_html' or self.mode == 'loose_html' ):
            end = [ '</body>', '</html>' ]
        else:
            end = [ ]

        return self.separator.join( self.header + self.content + self.footer + end )

    def __call__( self, escape=False ):
        """Return the document as a string.

        escape --   False   print normally
                    True    replace < and > by &lt; and &gt;
                            the default escape sequences in most browsers"""

        if escape:
            return _escape( self.__str__( ) )
        else:
            return self.__str__( )

    def add( self, text ):
        """This is an alias to addcontent."""
        self.addcontent( text )

    def addfooter( self, text ):
        """Add some text to the bottom of the document"""
        self.footer.append( text )

    def addheader( self, text ):
        """Add some text to the top of the document"""
        self.header.append( text )

    def addcontent( self, text ):
        """Add some text to the main part of the document"""
        self.content.append( text )


    def init( self, lang='en', css=None, metainfo=None, title=None, header=None,
              footer=None, charset=None, encoding=None, doctype=None, bodyattrs=None, script=None, base=None ):
        """This method is used for complete documents with appropriate
        doctype, encoding, title, etc information. For an HTML/XML snippet
        omit this method.

        lang --     language, usually a two character string, will appear
                    as <html lang='en'> in html mode (ignored in xml mode)
        
        css --      Cascading Style Sheet filename as a string or a list of
                    strings for multiple css files (ignored in xml mode)

        metainfo -- a dictionary in the form { 'name':'content' } to be inserted
                    into meta element(s) as <meta name='name' content='content'>
                    (ignored in xml mode)
        
        base     -- set the <base href="..."> tag in <head>
        
        bodyattrs --a dictionary in the form { 'key':'value', ... } which will be added
                    as attributes of the <body> element as <body key='value' ... >
                    (ignored in xml mode)

        script --   dictionary containing src:type pairs, <script type='text/type' src=src></script>
                    or a list of [ 'src1', 'src2', ... ] in which case 'javascript' is assumed for all

        title --    the title of the document as a string to be inserted into
                    a title element as <title>my title</title> (ignored in xml mode)

        header --   some text to be inserted right after the <body> element
                    (ignored in xml mode)

        footer --   some text to be inserted right before the </body> element
                    (ignored in xml mode)

        charset --  a string defining the character set, will be inserted into a
                    <meta http-equiv='Content-Type' content='text/html; charset=myset'>
                    element (ignored in xml mode)

        encoding -- a string defining the encoding, will be put into to first line of
                    the document as <?xml version='1.0' encoding='myencoding' ?> in
                    xml mode (ignored in html mode)

        doctype --  the document type string, defaults to
                    <!DOCTYPE HTML PUBLIC '-//W3C//DTD HTML 4.01 Transitional//EN'>
                    in html mode (ignored in xml mode)"""

        self._full = True

        if self.mode == 'strict_html' or self.mode == 'loose_html':
            if doctype is None:
                doctype = "<!DOCTYPE HTML PUBLIC '-//W3C//DTD HTML 4.01 Transitional//EN'>"
            self.header.append( doctype )
            self.html( lang=lang )
            self.head( )
            if charset is not None:
                self.meta( http_equiv='Content-Type', content="text/html; charset=%s" % charset )
            if metainfo is not None:
                self.metainfo( metainfo )
            if css is not None:
                self.css( css )
            if title is not None:
                self.title( title )
            if script is not None:
                self.scripts( script )
            if base is not None:
                self.base( href='%s' % base )
            self.head.close()
            if bodyattrs is not None:
                self.body( **bodyattrs )
            else:
                self.body( )
            if header is not None:
                self.content.append( header )
            if footer is not None:
                self.footer.append( footer )

        elif self.mode == 'xml':
            if doctype is None:
                if encoding is not None:
                    doctype = "<?xml version='1.0' encoding='%s' ?>" % encoding
                else:
                    doctype = "<?xml version='1.0' ?>"
            self.header.append( doctype )

    def css( self, filelist ):
        """This convenience function is only useful for html.
        It adds css stylesheet(s) to the document via the <link> element."""
      
        if isinstance( filelist, basestring ):
            self.link( href=filelist, rel='stylesheet', type='text/css', media='all' )
        else:
            for file in filelist:
                self.link( href=file, rel='stylesheet', type='text/css', media='all' )

    def metainfo( self, mydict ):
        """This convenience function is only useful for html.
        It adds meta information via the <meta> element, the argument is
        a dictionary of the form { 'name':'content' }."""

        if isinstance( mydict, dict ):
            for name, content in list( mydict.items( ) ):
                self.meta( name=name, content=content )
        else:
            raise TypeError( "Metainfo should be called with a dictionary argument of name:content pairs." )

    def scripts( self, mydict ):
        """Only useful in html, mydict is dictionary of src:type pairs or a list
        of script sources [ 'src1', 'src2', ... ] in which case 'javascript' is assumed for type.
        Will be rendered as <script type='text/type' src=src></script>"""

        if isinstance( mydict, dict ):
            for src, type in list( mydict.items( ) ):
                self.script( '', src=src, type='text/%s' % type )
        else:
            try:
                for src in mydict:
                    self.script( '', src=src, type='text/javascript' )
            except:
                raise TypeError( "Script should be given a dictionary of src:type pairs or a list of javascript src's." )


class _oneliner:
    """An instance of oneliner returns a string corresponding to one element.
    This class can be used to write 'oneliners' that return a string
    immediately so there is no need to instantiate the page class."""
    
    def __init__( self, case='lower' ):
        self.case = case
    
    def __getattr__( self, attr ):
        
        # tags should start with double underscore
        if attr.startswith("__") and attr.endswith("__"):
            raise AttributeError( attr )
        # tag with single underscore should be a reserved keyword
        if attr.startswith( '_' ):
            attr = attr.lstrip( '_' ) 
            if attr not in keyword.kwlist:
                raise AttributeError( attr )
        
        return element( attr, case=self.case, parent=None )

oneliner = _oneliner( case='lower' )
upper_oneliner = _oneliner( case='upper' )
given_oneliner = _oneliner( case='given' )

def _argsdicts( args, mydict ):
    """A utility generator that pads argument list and dictionary values, will only be called with len( args ) = 0, 1."""
    
    if len( args ) == 0:
        args = None, 
    elif len( args ) == 1:
        args = _totuple( args[0] )
    else:
        raise Exception( "We should have never gotten here." )

    mykeys = list( mydict.keys( ) )
    myvalues = list( map( _totuple, list( mydict.values( ) ) ) )

    maxlength = max( list( map( len, [ args ] + myvalues ) ) )

    for i in range( maxlength ):
        thisdict = { }
        for key, value in zip( mykeys, myvalues ):
            try:
                thisdict[ key ] = value[i]
            except IndexError:
                thisdict[ key ] = value[-1]
        try:
            thisarg = args[i]
        except IndexError:
            thisarg = args[-1]

        yield thisarg, thisdict

def _totuple( x ):
    """Utility stuff to convert string, int, long, float, None or anything to a usable tuple."""

    if isinstance( x, basestring ):
        out = x,
    elif isinstance( x, ( int, long, float ) ):
        out = str( x ),
    elif x is None:
        out = None,
    else:
        out = tuple( x )

    return out

def escape( text, newline=False ):
    """Escape special html characters."""

    if isinstance( text, basestring ):
        if '&' in text:
            text = text.replace( '&', '&amp;' )
        if '>' in text:
            text = text.replace( '>', '&gt;' )
        if '<' in text:
            text = text.replace( '<', '&lt;' )
        if '\"' in text:
            text = text.replace( '\"', '&quot;' )
        if '\'' in text:
            text = text.replace( '\'', '&quot;' )
        if newline:
            if '\n' in text:
                text = text.replace( '\n', '<br>' )

    return text

_escape = escape

def unescape( text ):
    """Inverse of escape."""
    
    if isinstance( text, basestring ):
        if '&amp;' in text:
            text = text.replace( '&amp;', '&' )
        if '&gt;' in text:
            text = text.replace( '&gt;', '>' )
        if '&lt;' in text:
            text = text.replace( '&lt;', '<' )
        if '&quot;' in text:
            text = text.replace( '&quot;', '\"' )

    return text

class dummy:
    """A dummy class for attaching attributes."""
    pass

doctype = dummy( )
doctype.frameset = """<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Frameset//EN" "http://www.w3.org/TR/html4/frameset.dtd">"""
doctype.strict = """<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">"""
doctype.loose = """<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">"""

class russell:
    """A dummy class that contains anything."""

    def __contains__( self, item ):
        return True


class MarkupError( Exception ):
    """All our exceptions subclass this."""
    def __str__( self ):
        return self.message

class ClosingError( MarkupError ):
    def __init__( self, tag ):
        self.message = "The element '%s' does not accept non-keyword arguments (has no closing tag)." % tag

class OpeningError( MarkupError ):
    def __init__( self, tag ):
        self.message = "The element '%s' can not be opened." % tag

class ArgumentError( MarkupError ):
    def __init__( self, tag ):
        self.message = "The element '%s' was called with more than one non-keyword argument." % tag

class InvalidElementError( MarkupError ):
    def __init__( self, tag, mode ):
        self.message = "The element '%s' is not valid for your mode '%s'." % ( tag, mode )

class DeprecationError( MarkupError ):
    def __init__( self, tag ):
        self.message = "The element '%s' is deprecated, instantiate markup.page with mode='loose_html' to allow it." % tag

class ModeError( MarkupError ):
    def __init__( self, mode ):
        self.message = "Mode '%s' is invalid, possible values: strict_html, html (alias for strict_html), loose_html, xml." % mode

class CustomizationError( MarkupError ):
    def __init__( self ):
        self.message = "If you customize the allowed elements, you must define both types 'onetags' and 'twotags'."

if __name__ == '__main__':
    import sys
    sys.stdout.write( __doc__ )
